/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#ifndef AIRCRAFT_H_
#define AIRCRAFT_H_
#include <stdint.h>
#include <string>
#include <vector>
#include "WBEnvelope.h"

#define AIRCRAFT_FILEMAGIC "Vifir_Aircraft"

/* Although this information is all displayed in the most common
 * units in the US, it is all stored in SI units.  This is for 2
 * reasons:
 * 1. The author prefers SI units in general
 * 2. It forces conversion routines to be created at each step
 *    so that none are missed.
 */
class Aircraft {
public:
	/* The INVALID enum is used for bounds checking - it should always be the last entry */
	enum FuelRateModelEnum {
		FUELRATEMODEL_SEGMENT, FUELRATEMODEL_HOURLY, FUELRATEMODEL_INVALID
	};
	const static unsigned int FUELRATEMODEL_HOURLY_NHOURS=10;

	enum AirspeedUnitsEnum {
		AIRSPEEDUNITS_KPH,
		AIRSPEEDUNITS_MPH,
		AIRSPEEDUNITS_KNOTS,
		AIRSPEEDUNITS_INVALID
	};
	const static char AirspeedUnitsText[][10];

	enum ClimbDescentRateUnitsEnum {
		CLIMBDESCENTRATEUNITS_MPERS,
		CLIMBDESCENTRATEUNITS_FTPERMIN,
		CLIMBDESCENTRATEUNITS_KNOTS,
		CLIMBDESCENTRATEUNITS_INVALID
	};
	const static char ClimbDescentRateUnitsText[][20];

	enum AltitudeUnitsEnum {
		ALTITUDEUNITS_METERS,
		ALTITUDEUNITS_FEET,
		ALTITUDEUNITS_INVALID
	};
	const static char AltitudeUnitsText[][10];

	enum AircraftWeightUnitsEnum {
		AIRCRAFTWEIGHTUNITS_KG,
		AIRCRAFTWEIGHTUNITS_LB,
		AIRCRAFTWEIGHTUNITS_INVALID
	};
	const static char AircraftWeightUnitsText[][10];

	enum CGStationDistUnitsEnum {
		CGSTATIONDISTUNITS_M,
		CGSTATIONDISTUNITS_CM,
		CGSTATIONDISTUNITS_FT,
		CGSTATIONDISTUNITS_IN,
		CGSTATIONDISTUNITS_INVALID
	};
	const static char CGStationDistUnitsText[][20];

	enum OilQtyUnitsEnum {
		OILQTYUNITS_KG,
		OILQTYUNITS_L,
		OILQTYUNITS_LB,
		OILQTYUNITS_GAL,
		OILQTYUNITS_QT,
		OILQTYUNITS_INVALID
	};
	const static char OilQtyUnitsText[][10];

	enum FuelQtyUnitsEnum {
		FUELQTYUNITS_KG,
		FUELQTYUNITS_L,
		FUELQTYUNITS_LB,
		FUELQTYUNITS_GAL,
		FUELQTYUNITS_INVALID
	};
	const static char FuelQtyUnitsText[][10];

	enum FuelTypeEnum {
		FUELTYPE_AVGAS,
		FUELTYPE_MOGAS,
		FUELTYPE_JET,
		FUELTYPE_INVALID
	};
	const static char FuelTypeText[][10];
	const static double FuelTypeDensity[]; /* kg/L, one for each FuelTypeEnum */

	Aircraft();
	~Aircraft();

	std::string GetDesc();
	void SetDesc(std::string value);
	std::string GetRegistration();
	void SetRegistration(std::string value);
	double GetTASClimb();
	void SetTASClimb(double value);
	double GetTASCruise();
	void SetTASCruise(double value);
	double GetTASDescent();
	void SetTASDescent(double value);
	double GetClimbRate();
	void SetClimbRate(double value);
	double GetDescentRate();
	void SetDescentRate(double value);
	enum FuelRateModelEnum GetFuelRateModel();
	void SetFuelRateModel(enum FuelRateModelEnum value);
	double GetFuelRateClimb();
	void SetFuelRateClimb(double value);
	double GetFuelRateCruise();
	void SetFuelRateCruise(double value);
	double GetFuelRateDescent();
	void SetFuelRateDescent(double value);
	double GetFuelRateHourly(unsigned int Hour);
	void SetFuelRateHourly(double value, unsigned int Hour);
	enum AirspeedUnitsEnum GetAirspeedUnits();
	void SetAirspeedUnits(enum AirspeedUnitsEnum);
	enum ClimbDescentRateUnitsEnum GetClimbDescentRateUnits();
	void SetClimbDescentRateUnits(enum ClimbDescentRateUnitsEnum);
	enum AltitudeUnitsEnum GetAltitudeUnits();
	void SetAltitudeUnits(enum AltitudeUnitsEnum);
	enum AircraftWeightUnitsEnum GetAircraftWeightUnits();
	void SetAircraftWeightUnits(enum AircraftWeightUnitsEnum);
	enum CGStationDistUnitsEnum GetCGStationDistUnits();
	void SetCGStationDistUnits(enum CGStationDistUnitsEnum);
	enum OilQtyUnitsEnum GetOilQtyUnits();
	void SetOilQtyUnits(enum OilQtyUnitsEnum);
	enum FuelQtyUnitsEnum GetFuelQtyUnits();
	void SetFuelQtyUnits(enum FuelQtyUnitsEnum);
	enum FuelTypeEnum GetFuelType();
	void SetFuelType(enum FuelTypeEnum value);
	std::string GetFilename();
	bool Load(std::string Filename);
	bool Load();
	void Save(std::string Filename);
	void Save();
	bool MakeBinary(struct Aircraft_Binary_struct *BinaryData);
	void SetFromBinary(struct Aircraft_Binary_struct *BinaryData);

private:
	std::string Filename;
	std::string Desc;
	std::string Registration;

	double TASClimb, TASCruise, TASDescent; /* km/h */
	double ClimbRate, DescentRate; /* m/s */
	enum FuelRateModelEnum FuelRateModel;
	union {
		struct {
			double Climb, Cruise, Descent; /* L/h */
		} Segment; /* Segment-based for piston aircraft */
		double Hour[FUELRATEMODEL_HOURLY_NHOURS]; /* kg/h */
	} FuelRate;

	enum AirspeedUnitsEnum AirspeedUnits;
	enum ClimbDescentRateUnitsEnum ClimbDescentRateUnits;
	enum AltitudeUnitsEnum AltitudeUnits;
	enum AircraftWeightUnitsEnum AircraftWeightUnits;
	enum CGStationDistUnitsEnum CGStationDistUnits;
	enum OilQtyUnitsEnum OilQtyUnits;
	enum FuelQtyUnitsEnum FuelQtyUnits;
	enum FuelTypeEnum FuelType;

	std::vector<WBEnvelope*> WBEnvelopes;
};

class WBStation {
public:
	enum StationTypeEnum {
		STATIONTYPE_BEW,STATIONTYPE_PAYLOAD,STATIONTYPE_FUEL,STATIONTYPE_INVALID
	};

	WBStation();

	void SetByArm(enum StationTypeEnum Type, double Mass, double Arm);
	void SetByMom(enum StationTypeEnum Type, double Mass, double Moment);
	double GetMass();
	double GetArm();
	double GetMoment();

private:
	double Mass; /* kg */
	double Arm; /* m */
	enum StationTypeEnum Type;
};

/* Increment the version when the structure changes */
#define AIRCRAFT_BINARY_STRUCT_VER 1
#ifdef WIN32
#pragma pack(1)
#endif
struct Aircraft_Binary_struct {
	char Magic[sizeof(AIRCRAFT_FILEMAGIC)];
	uint32_t DataVersion;
	uint32_t CRC32;
	char Desc[128];
	char Registration[16];
	uint32_t TASClimbx10000, TASCruisex10000, TASDescentx10000;
	uint32_t ClimbRatex10000, DescentRatex10000;
	uint8_t FuelRateModel;
	uint32_t FuelRateClimbx10000, FuelRateCruisex10000, FuelRateDescentx10000;
	uint32_t FuelRateHourx10000[Aircraft::FUELRATEMODEL_HOURLY_NHOURS];
	uint8_t AirspeedUnits;
	uint8_t ClimbDescentRateUnits;
	uint8_t AltitudeUnits;
	uint8_t AircraftWeightUnits;
	uint8_t CGStationDistUnits;
	uint8_t OilQtyUnits;
	uint8_t FuelQtyUnits;
	uint8_t FuelType;

	uint8_t nWBEnvelopes; /* Add check to prevent overflow */
	/* Add envelopes here */

}
#ifndef WIN32
        __attribute__ ((packed))
#endif
;
#ifdef WIN32
#pragma pack()
#endif

#endif /*AIRCRAFT_H_*/
