/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <cstring>
#include <sstream>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <libglademm.h>
#include <glibmm/i18n.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/main.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/stock.h>
#include "ImportNationalFlightDatabase.h"
#include "Exception.h"
#include "Util.h"
#include "DataStore.h"

ImportNationalFlightDatabase::ImportNationalFlightDatabase(Config *Cfg, std::string *GladeFilename, DataStore *Data)
{
	this->Cfg=Cfg;
	this->GladeFilename=GladeFilename;
	this->Data=Data;

	try
	{
		std::string gladefile;

		refXml = Gnome::Glade::Xml::create(*GladeFilename, "ProgressWindow");
	}
	catch (const Gnome::Glade::XmlError& ex)
	{
		Gtk::MessageDialog *message = new Gtk::MessageDialog(_("Error reading data file: ")+*GladeFilename+"\n"+_("Program cannot continue."),
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		Gtk::Main::quit();
		return;
	}

	refXml->get_widget("ProgressWindow",ProgressWindow);
	refXml->get_widget("SubProgressBar",SubProgressBar);
	refXml->get_widget("ProgressBar",ProgressBar);
	refXml->get_widget("CancelButton",CancelButton);

	ProgressWindow->set_deletable(false);
	CancelButton->signal_clicked().connect(sigc::mem_fun(this,&ImportNationalFlightDatabase::on_CancelButton_clicked));

	CancelFlag=false;
	BytesRead=0;
	BytesAvail=0;
	gettimeofday(&LastProgressBarUpdate,NULL);
}

ImportNationalFlightDatabase::~ImportNationalFlightDatabase() {
	delete ProgressWindow;
}

int ImportNationalFlightDatabase::Run()
{
	std::vector<std::string> *Lines;
	std::vector<std::string>::iterator Linesi;
	int Count;
	struct stat statdata;

	Gtk::FileChooserDialog FileDialog("Browse to the NFD data", Gtk::FILE_CHOOSER_ACTION_OPEN);
	FileDialog.set_current_folder(Cfg->GetNationalFlightDBImportDir());

	/* Add filters */
	Gtk::FileFilter FileFilter;
	FileFilter.set_name("National Flight Database (FAANFD18)");
	FileFilter.add_pattern("FAANFD18");
	FileDialog.add_filter(FileFilter);

	/* Add response buttons */
	FileDialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	FileDialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	if (FileDialog.run()!=Gtk::RESPONSE_OK)
		return false;

	Cfg->SetNationalFlightDBImportDir(FileDialog.get_current_folder());

	if (stat(FileDialog.get_filename().c_str(),&statdata)!=0) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			"Error checking NFD file.",false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		return false;
	}
	BytesAvail=statdata.st_size;

	/* It's actually a text file, but we handle it as binary so that Win32 handles it the same
	 * way as everybody else.
	 */
	IFStream.open(FileDialog.get_filename().c_str(),std::ifstream::in | std::ifstream::binary);
	if (!IFStream.good()) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			"Error opening NFD file.",false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		return false;
	}
	FileDialog.hide();

	/* Read data from the file */
	if (!NFDData.ParseHeader(&IFStream)) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			"Error parsing NFD header.",false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		return false;
	}

	if (!ConfirmData())
		return false;

	std::string VolName=NFDData.GetVolume();
	time_t EffFrom=Volume::ComputeEffectiveFrom(NFDData.GetVolume());
	time_t EffTo=Volume::ComputeEffectiveTo(NFDData.GetVolume());
	VolData=Data->GetVolume(&VolName,&EffFrom,&EffTo);
	if (!VolData) {
		throw Exception("Error creating store for volume data",Exception::LEVEL_ERROR);
	}

	if (VolData->GetnApproachPlates()) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			"This will remove Approach Plate data (but not the image files) for this Volume - it will need to be re-imported.",false,Gtk::MESSAGE_WARNING,
			Gtk::BUTTONS_OK_CANCEL,true);
		if (message->run()!=Gtk::RESPONSE_OK) {
			delete message;
			return false;
		}
		delete message;
	}

	SubProgressBar->set_text("Reading");
	ProgressBar->set_text("Total");
	ProgressWindow->show();
	ProgressDone=0.0;

	NFDData.ReserveLines();

	Count=0;
	while (ReadLine() && !CancelFlag) {
		if (Count%1000==0) {
			ScheduleMainLoop();
		}
		Count++;
	}

	IFStream.close();
	ProgressDone+=ReadingProgressFrac;

	/* Clear the existing datastore */
	VolData->ClearFlightDB();

	/* Parse the airport data */
	if (CancelFlag) {
		NFDData.FreeLines();
		return false;
	}

	SubProgressBar->set_text("Parsing Airports");
	Count=0;

	Lines=NFDData.GetAirportLines();
	for (Linesi=(*Lines).begin();Linesi!=(*Lines).end() && !CancelFlag;Linesi++) {
		SubProgressBar->set_fraction((double)Count/(*Lines).size());
		ProgressBar->set_fraction(ProgressDone+ParseAirportsProgressFrac*Count/(*Lines).size());
		NFDData.ParseAirportLine(Linesi,VolData->GetAirports());
		ScheduleMainLoop();
		Count++;
	}
	ProgressDone+=ParseAirportsProgressFrac;

	/* Parse the runway data */
	if (CancelFlag) {
		NFDData.FreeLines();
		return false;
	}

	SubProgressBar->set_text("Parsing Runways");
	Count=0;

	Lines=NFDData.GetRunwayLines();
	for (Linesi=(*Lines).begin();Linesi!=(*Lines).end() && !CancelFlag;Linesi++) {
		SubProgressBar->set_fraction((double)Count/(*Lines).size());
		ProgressBar->set_fraction(ProgressDone+ParseRunwaysProgressFrac*Count/(*Lines).size());
		NFDData.ParseRunwayLine(Linesi,VolData->GetAirports());
		ScheduleMainLoop();
		Count++;
	}
	ProgressDone+=ParseRunwaysProgressFrac;

	/* Parse the VOR data */
	if (CancelFlag) {
		NFDData.FreeLines();
		return false;
	}

	SubProgressBar->set_text("Parsing VORs");
	Count=0;

	Lines=NFDData.GetVORLines();
	for (Linesi=(*Lines).begin();Linesi!=(*Lines).end() && !CancelFlag;Linesi++) {
		SubProgressBar->set_fraction((double)Count/(*Lines).size());
		ProgressBar->set_fraction(ProgressDone+ParseVORsProgressFrac*Count/(*Lines).size());
		NFDData.ParseVORLine(Linesi,VolData->GetNavAids());
		ScheduleMainLoop();
		Count++;
	}
	ProgressDone+=ParseVORsProgressFrac;

	/* Parse the NDB data */
	if (CancelFlag) {
		NFDData.FreeLines();
		return false;
	}

	SubProgressBar->set_text("Parsing NDBs");
	Count=0;

	Lines=NFDData.GetNDBLines();
	for (Linesi=(*Lines).begin();Linesi!=(*Lines).end() && !CancelFlag;Linesi++) {
		SubProgressBar->set_fraction((double)Count/(*Lines).size());
		ProgressBar->set_fraction(ProgressDone+ParseNDBsProgressFrac*Count/(*Lines).size());
		NFDData.ParseNDBLine(Linesi,VolData->GetNavAids());
		ScheduleMainLoop();
		Count++;
	}
	ProgressDone+=ParseNDBsProgressFrac;

	/* Parse the Enroute Waypoint data */
	if (CancelFlag) {
		NFDData.FreeLines();
		return false;
	}

	SubProgressBar->set_text("Parsing Waypoints");
	Count=0;

	Lines=NFDData.GetEnrouteWaypointLines();
	for (Linesi=(*Lines).begin();Linesi!=(*Lines).end() && !CancelFlag;Linesi++) {
		SubProgressBar->set_fraction((double)Count/(*Lines).size());
		NFDData.ParseEnrouteWaypointLine(Linesi,VolData->GetEnrouteWaypoints());
		ProgressBar->set_fraction(ProgressDone+ParseWaypointsProgressFrac*Count/(*Lines).size());
		ScheduleMainLoop();
		Count++;
	}

	/* The lines of data have all been parsed - free the memory holding them. */
	NFDData.FreeLines();

	std::string Filename;
	Filename=Cfg->GetDataStoreDir()+"/datastore.vifir";
	Util::ShellExpand(&Filename);
	try {
		VolData->Save();
	}
	catch (Exception& Ex) {
		Ex.Show();
		return false;
	}
	return true;
}

bool ImportNationalFlightDatabase::ReadLine() {
	char buf[256];

	IFStream.getline(buf,sizeof(buf));
	if (IFStream.good()) {
		NFDData.StoreLine(buf);
		BytesRead+=strlen(buf);
		SubProgressBar->set_fraction((double)BytesRead/BytesAvail);
		ProgressBar->set_fraction(ReadingProgressFrac*BytesRead/BytesAvail);
		ScheduleMainLoop();
		return true;
	} else {
		if (!IFStream.eof()) {
			Gtk::MessageDialog *message = new Gtk::MessageDialog(
				"Error reading NFD file.",false,Gtk::MESSAGE_ERROR,
				Gtk::BUTTONS_CANCEL,true);
			message->run();
			delete message;
			CancelFlag=true;
		}
		return false;
	}
	return false;
}

void ImportNationalFlightDatabase::on_CancelButton_clicked() {
	CancelFlag=true;
}

/* Allow the main loop to run (and the GUI to update from time to time */
void ImportNationalFlightDatabase::ScheduleMainLoop() {
	struct timeval CurrTime;
	double Last,Curr;

	Last=LastProgressBarUpdate.tv_sec+LastProgressBarUpdate.tv_usec/1000000.0;
	gettimeofday(&CurrTime,NULL);
	Curr=CurrTime.tv_sec+CurrTime.tv_usec/1000000.0;
	if (Curr-Last>0.01) {
		Gtk::Main::iteration(false);
		gettimeofday(&LastProgressBarUpdate,NULL);
	}
}

bool ImportNationalFlightDatabase::ConfirmData() {
	struct tm bdtime;
	time_t Timestamp;
	std::string VolumeStr;
	std::ostringstream EffFromOStr,EffToOStr;
	bool result=false;

	VolumeStr="Volume="+NFDData.GetVolume()+"\n";
	Timestamp=Volume::ComputeEffectiveFrom(NFDData.GetVolume());

	try {
		Util::GmTime_r(&bdtime,Timestamp);
	}
	catch (Exception& e) {
		e.Show();
		return false;
	}

	EffFromOStr<<"Eff From="<<std::setfill('0')<<
		std::setw(2)<<bdtime.tm_hour<<std::setw(2)<<bdtime.tm_min<<"Z "<<
		std::setw(2)<<(bdtime.tm_mon+1)<<"/"<<
		std::setw(2)<<bdtime.tm_mday<<"/"<<
		std::setw(2)<<(bdtime.tm_year-100)<<"\n";
	Timestamp=Volume::ComputeEffectiveTo(NFDData.GetVolume())+1;

	try {
		Util::GmTime_r(&bdtime,Timestamp);
	}
	catch (Exception& e) {
		e.Show();
		return false;
	}

	EffToOStr<<"Eff To="<<std::setfill('0')<<
		std::setw(2)<<bdtime.tm_hour<<std::setw(2)<<bdtime.tm_min<<"Z "<<
		std::setw(2)<<(bdtime.tm_mon+1)<<"/"<<
		std::setw(2)<<bdtime.tm_mday<<"/"<<
		std::setw(2)<<(bdtime.tm_year-100)<<"\n";


	Gtk::MessageDialog *message = new Gtk::MessageDialog(
		VolumeStr+EffFromOStr.str()+EffToOStr.str(),false,Gtk::MESSAGE_QUESTION,
		Gtk::BUTTONS_OK_CANCEL,true);
	message->set_title("Confirm");

	result=(message->run()==Gtk::RESPONSE_OK);
	delete message;

	return result;
}
