/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#ifndef WIN32
#include <wordexp.h>
#endif
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <cstring>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <cctype>
#include <memory>
#include <gtkmm/main.h>
#include "Util.h"
#include "Exception.h"

#ifdef WIN32
	#include <windows.h>
#endif

void Util::ShellExpand(std::string *str) {
#ifndef WIN32
	int result;
	unsigned int i;
	wordexp_t we;

	if ((result=wordexp(str->c_str(),&we,0))==0) {
		(*str)="";
		for (i=0;i<we.we_wordc;i++) {
			if (i!=0)
				(*str)+=" ";
			(*str)+=we.we_wordv[i];
		}
		wordfree(&we);
	}
#else
	/* This is fairly brittle and only handles paths beginning with '~' */
	if ((*str)[0]!='~')
		return;

	*str=str->substr(1);	/* Strip off the leading '~' */
	char *env=getenv("USERPROFILE");
	if (env) {
		/* Replace the '~' with the user's home directory */
		*str=std::string(env)+"\\Local Settings\\Application Data"+(*str);
	}
	std::replace(str->begin(), str->end(), '/', '\\');
#endif
}

void Util::ChkDir(std::string DirPath) {
	/* Create the directory if it doesn't exist */
	struct stat statdata;
	std::ostringstream Text;

	ShellExpand(&DirPath);
	errno=0;
	if (stat(DirPath.c_str(),&statdata)!=0) {
		if (errno==ENOENT) {
			if (Util::MkDir(DirPath,0755)==0) {
				std::cout<<"Created "<<DirPath<<"."<<std::endl;
			} else {
				Text<<"Error creating "<<DirPath<<": "<<Util::StrError()<<" ("<<errno<<")";
				throw Exception(Text.str(),Exception::LEVEL_ERROR);
			}
		}
	}
}

/* Return time based on Greenwich Mean Time */
time_t Util::GMMkTime(struct tm *tm) {
	time_t time;

#ifndef WIN32
	char *oldTZ;

	oldTZ=getenv("TZ");
	setenv("TZ","GMT",1);
	tzset();
	time=mktime(tm);

	if (oldTZ)
		setenv("TZ",oldTZ,1);
	else
		unsetenv("TZ");
	tzset();
	return time;
#else
	time=mktime(tm)-_timezone;
	return time;
#endif
}

/* gmtime_r is not available on Win32, so we basically reimplement it here */
void Util::GmTime_r(struct tm *bdtime, time_t Timestamp) {
	struct tm *tmpbd;

	errno=0;
	tmpbd=gmtime(&Timestamp);

	if (!tmpbd) {
		throw Exception(std::string("Error parsing time: ")+Util::StrError(),Exception::LEVEL_ERROR);
	}
	memcpy(bdtime,tmpbd,sizeof(*bdtime));
}

bool Util::SmartCopyFile(std::string SrcPath, std::string DestDir, std::string DestName, OverwriteFileDialog *OFD) {
	std::string DstPath,ErrStr;
	std::ifstream ifs;
	std::ofstream ofs;
	struct stat statdata;
	char *buffer=NULL;
	const int bufsize=1024*1024;

	if (DestName.size()) {
		DstPath=DestDir+DestName;
	} else {
		DstPath=DestDir+SrcPath.substr(SrcPath.find_last_of("/\\")+1,SrcPath.size());
	}

	buffer=(char*)malloc(bufsize);
	if (!buffer) {
		throw Exception("Error allocating memory",Exception::LEVEL_ERROR);
	}

	if (OFD) {
		if (stat(DstPath.c_str(),&statdata)==0) {
			/* File exists */
			switch (OFD->Run(DstPath)) {
				case OverwriteFileDialog::RESPONSE_CANCEL:
					free(buffer);
					throw Exception("Cancelled",Exception::LEVEL_WARNING);
					break;
				case OverwriteFileDialog::RESPONSE_KEEP:
					free(buffer);
					return true;	/* As if we were successful */
				case OverwriteFileDialog::RESPONSE_OVERWRITE:
					break;			/* Fall through as if nothing happened */
			}
		}
	}

	ifs.open(SrcPath.c_str(),std::ios::binary);
	if (!ifs.good()) {
		ErrStr=Util::StrError();
		free(buffer);
		throw Exception("Error opening "+SrcPath+": "+ErrStr,Exception::LEVEL_ERROR);
	}

	ofs.open(DstPath.c_str(),std::ios::binary);
	if (!ofs.good()) {
		ErrStr=Util::StrError();
		ifs.close();
		free(buffer);
		throw Exception("Error opening "+DstPath+": "+ErrStr,Exception::LEVEL_ERROR);
	}

	do {
		ifs.read(buffer,bufsize);
		ofs.write(buffer,ifs.gcount());
		Util::ScheduleMainLoop();
	} while (ifs.good() && ofs.good());

	if (!ofs.good()) {
		ErrStr=Util::StrError();
		ifs.close();
		ofs.close();
		free(buffer);
		throw Exception("Error writing to "+DstPath+": "+ErrStr,Exception::LEVEL_ERROR);
	}

	if (!ifs.good() && !ifs.eof()) {
		ErrStr=Util::StrError();
		ifs.close();
		ofs.close();
		free(buffer);
		throw Exception("Error reading from "+SrcPath+": "+ErrStr,Exception::LEVEL_ERROR);
	}

	ifs.close();
	ofs.close();
	free(buffer);

	return true;
}

/* Allow the main loop to run (and the GUI to update from time to time */
void Util::ScheduleMainLoop() {
	struct timeval CurrTime;
	double Last,Curr;
	static struct timeval LastTV;
	static bool firsttime=true;

	if (firsttime) {
		LastTV.tv_sec=0;
		LastTV.tv_usec=0;
		firsttime=false;
	}

	Last=LastTV.tv_sec+LastTV.tv_usec/1000000.0;
	gettimeofday(&CurrTime,NULL);
	Curr=CurrTime.tv_sec+CurrTime.tv_usec/1000000.0;
	if (Curr-Last>0.1) {
		while (Gtk::Main::events_pending())
			Gtk::Main::iteration(false);
		gettimeofday(&LastTV,NULL);
	}
}

std::string Util::toUpper(std::string str) {
	for (unsigned int i=0;i<str.length();i++) {
		str[i] = toupper(str[i]);
	}
	return str;
}

std::string Util::DoubleToStr(const double Number) {
	std::ostringstream ostr;

	if (!(ostr<<std::setprecision(5)<<std::setw(0)<<Number))
		return "";
	return ostr.str();
}

void Util::Explode(const std::string Str, const std::string Delim, std::vector<std::string> *Tokens) {
	size_t Start=0, End=0;

	Tokens->clear();

	End=Str.find_first_of(Delim,Start);
	while (End!=std::string::npos) {
		std::auto_ptr<std::string> Token(new std::string);
		*Token=Str.substr(Start,End-Start);
		Tokens->push_back(*Token);
		Start=End+1;
		End=Str.find_first_of(Delim,Start);
	}
}

int Util::MkDir(std::string Path, int Mode) {
	int result;

	errno=0;
#ifndef WIN32
	result=mkdir(Path.c_str(),Mode);
#else
	result=mkdir(Path.c_str());
#endif

	if (result!=0 && errno!=EEXIST) {
		Exception("Error creating directory "+Path+": "+Util::StrError(),Exception::LEVEL_ERROR).Show();
	}

	return result;
}

char *Util::StrError() {
#ifndef WIN32
	return strerror(errno);
#else
	static char buf[1024];
	DWORD result;

	result=FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,0,GetLastError(),0,buf,sizeof(buf),NULL);

	if (result==0) {
		strncpy(buf,"Unknown error.",sizeof(buf));
	}
	return buf;
#endif
}

void Util::LonLatToDMMSSPSSSS(double Lon, double Lat, std::string *str) {
	std::ostringstream Text;
	std::string NS,EW;
	int LatDeg,LonDeg,LatMin,LonMin;
	double LatSec,LonSec;

	if (Lat<0)
		NS="S";
	else
		NS="N";
	if (Lon<0)
		EW="W";
	else
		EW="E";
	Lat=fabs(Lat);
	Lon=fabs(Lon);

	LatDeg=(int)(Lat);
	LonDeg=(int)(Lon);

	Lat-=LatDeg;
	Lon-=LonDeg;

	LatMin=(int)(Lat*60.0);
	LonMin=(int)(Lon*60.0);

	Lat-=LatMin/60.0;
	Lon-=LonMin/60.0;

	LatSec=Lat*3600.0;
	LonSec=Lon*3600.0;

	Text.setf(std::ios::fixed);
	Text<<
		std::setprecision(0)<<std::setw(0)<<LatDeg<<"° "<<
		std::setprecision(0)<<std::setw(2)<<std::setfill('0')<<LatMin<<"´ "<<
		std::setprecision(4)<<std::setw(4+3)<<std::setfill('0')<<LatSec<<"˝ "<<
		NS<<", "<<
		std::setprecision(0)<<std::setw(0)<<LonDeg<<"° "<<
		std::setprecision(0)<<std::setw(2)<<std::setfill('0')<<LonMin<<"´ "<<
		std::setprecision(4)<<std::setw(4+3)<<std::setfill('0')<<LonSec<<"˝ "<<
		EW;

	*str=Text.str();
}

void Util::LonLatToDMMPMMMMMM(double Lon, double Lat, std::string *str) {
	std::ostringstream Text;
	std::string NS,EW;
	int LatDeg,LonDeg;
	double LatMin,LonMin;

	if (Lat<0)
		NS="S";
	else
		NS="N";
	if (Lon<0)
		EW="W";
	else
		EW="E";
	Lat=fabs(Lat);
	Lon=fabs(Lon);

	LatDeg=(int)(Lat);
	LonDeg=(int)(Lon);

	Lat-=LatDeg;
	Lon-=LonDeg;

	LatMin=Lat*60.0;
	LonMin=Lon*60.0;

	Text.setf(std::ios::fixed);
	Text<<
		std::setprecision(0)<<std::setw(0)<<LatDeg<<"° "<<
		std::setprecision(6)<<std::setw(6+3)<<std::setfill('0')<<LatMin<<"´ "<<
		NS<<", "<<
		std::setprecision(0)<<std::setw(0)<<LonDeg<<"° "<<
		std::setprecision(6)<<std::setw(6+3)<<std::setfill('0')<<LonMin<<"´ "<<
		EW;

	*str=Text.str();
}

void Util::LonLatToDPDDDDDDD(double Lat, double Lon, std::string *str) {
	std::ostringstream Text;
	Text.setf(std::ios::fixed);
	Text<<std::setprecision(7)<<Lon<<","<<Lat;
	*str=Text.str();
}

std::string Util::BDTimeToMM_DD_HH_MM(struct tm *bdtime) {
	std::stringstream SStr;
	SStr<<std::setw(2)<<std::setfill('0');
	SStr<<bdtime->tm_mon+1<<"/";
	SStr<<std::setw(2)<<std::setfill('0');
	SStr<<bdtime->tm_mday<<" ";
	SStr<<std::setw(2)<<std::setfill('0');
	SStr<<bdtime->tm_hour<<":";
	SStr<<std::setw(2)<<std::setfill('0');
	SStr<<bdtime->tm_min;
	return SStr.str();
}

time_t Util::CalcEffectiveFrom(std::string *Volume) {
	int EffYear, Edition, Editioni=1, PrevYear=2000;
	time_t Timestamp;
	struct tm bdtime;

	if (Volume->size()!=4)
		return 0;
	EffYear=atoi(Volume->substr(0,2).c_str());
	if (EffYear<=0 || EffYear>99)
		return 0;
	Edition=atoi(Volume->substr(2,2).c_str());
	if (Edition<=0 || Edition>13)
		return 0;

	/* Initialize bdtime */
	Timestamp=time(NULL);
	Util::GmTime_r(&bdtime,Timestamp);

	/* Set the start date/time - assume Volume 0000 became effective Jan 27, 2000 */
	bdtime.tm_year=100;     /* Year 2000 */
	bdtime.tm_mon=0;
	bdtime.tm_mday=27;
	bdtime.tm_hour=9;
	bdtime.tm_min=1;
	bdtime.tm_sec=0;

	while (!(bdtime.tm_year-100==EffYear && Editioni==Edition) && bdtime.tm_year!=199) {
		Timestamp=GMMkTime(&bdtime)+28*24*60*60;        /* Go forward 28 days */
		Util::GmTime_r(&bdtime,Timestamp);
		Editioni++;
		if (bdtime.tm_year!=PrevYear-1900) {
			PrevYear=bdtime.tm_year+1900;
			Editioni=1;
		}
	}

	if (bdtime.tm_year==199)        /* No suitable cycle was found */
		return 0;

	return Timestamp;
}

time_t Util::CalcEffectiveTo(std::string *VolName) {
	/* Valid for 28 days.  Subtract 1 second so there's no overlap */
	return CalcEffectiveFrom(VolName)+28*24*60*60-1;
}

